/* Split-Radix FFT.c */
/* 2011.06.20, 10RU004, ΋ˑ */

#include "Split-Radix FFT.h"

void FFT_main(long n, double ar[], double ai[])
{
	long m, mq, irev, i = 0, j0, j1, j2, j3, k;
	double theta, w1r, w1i, w3r, w3i;
	double x0r, x0i, x1r, x1i, x3r, x3i;

	for (j0 = 1; j0 < n - 1; j0++)
	{
		for (k = n >> 1; k > (i ^= k); k >>= 1);
		if (j0 < i){
			x0r = ar[j0];
			x0i = ai[j0];
			ar[j0] = ar[i];
			ai[j0] = ai[i];
			ar[i] = x0r;
			ai[i] = x0i;
		}
	}
	theta = -2 * atan(1.0) / n;

	for (m = 4; m <= n; m <<= 1)
	{
		mq = m >> 2;

		for (k = mq; k >= 1; k >>= 2)
		{
			for (j0 = mq - k; j0 < mq - (k >> 1); j0++)
			{
				j1 = j0 + mq;
				j2 = j1 + mq;
				j3 = j2 + mq;
				x1r = ar[j0] - ar[j1];
				x1i = ai[j0] - ai[j1];
				ar[j0] += ar[j1];
				ai[j0] += ai[j1];
				x3r = ar[j3] - ar[j2];
				x3i = ai[j3] - ai[j2];
				ar[j2] += ar[j3];
				ai[j2] += ai[j3];
				ar[j1] = x1r - x3i;
				ai[j1] = x1i + x3r;
				ar[j3] = x1r + x3i;
				ai[j3] = x1i - x3r;
			}
		}
		if (m == n) continue;

		irev = n >> 1;

		w1r = cos(theta * irev);

		for (k = mq; k >= 1; k >>= 2)
		{
			for (j0 = m + mq - k; j0 < m + mq - (k >> 1); j0++)
			{
				j1 = j0 + mq;
				j2 = j1 + mq;
				j3 = j2 + mq;
				x1r = ar[j0] - ar[j1];
				x1i = ai[j0] - ai[j1];
				ar[j0] += ar[j1];
				ai[j0] += ai[j1];
				x3r = ar[j3] - ar[j2];
				x3i = ai[j3] - ai[j2];
				ar[j2] += ar[j3];
				ai[j2] += ai[j3];
				x0r = x1r - x3i;
				x0i = x1i + x3r;
				ar[j1] = w1r * (x0r + x0i);
				ai[j1] = w1r * (x0i - x0r);
				x0r = x1r + x3i;
				x0i = x1i - x3r;
				ar[j3] = w1r * (-x0r + x0i);
				ai[j3] = w1r * (-x0i - x0r);
			}
		}

		for (i = 2 * m; i < n; i += m)
		{
			for (k = n >> 1; k > (irev ^= k); k >>= 1);

			w1r = cos(theta * irev);
			w1i = sin(theta * irev);
			w3r = cos(theta * 3 * irev);
			w3i = sin(theta * 3 * irev);

			for (k = mq; k >= 1; k >>= 2)
			{
				for (j0 = i + mq - k; j0 < i + mq - (k >> 1); j0++)
				{
					j1 = j0 + mq;
					j2 = j1 + mq;
					j3 = j2 + mq;
					x1r = ar[j0] - ar[j1];
					x1i = ai[j0] - ai[j1];
					ar[j0] += ar[j1];
					ai[j0] += ai[j1];
					x3r = ar[j3] - ar[j2];
					x3i = ai[j3] - ai[j2];
					ar[j2] += ar[j3];
					ai[j2] += ai[j3];
					x0r = x1r - x3i;
					x0i = x1i + x3r;
					ar[j1] = w1r * x0r - w1i * x0i;
					ai[j1] = w1r * x0i + w1i * x0r;
					x0r = x1r + x3i;
					x0i = x1i - x3r;
					ar[j3] = w3r * x0r - w3i * x0i;
					ai[j3] = w3r * x0i + w3i * x0r;
				}
			}
		}
	}
	mq = n >> 1;

	for (k = mq; k >= 1; k >>= 2)
	{
		for (j0 = mq - k; j0 < mq - (k >> 1); j0++)
		{
			j1 = mq + j0;
			x0r = ar[j0] - ar[j1];
			x0i = ai[j0] - ai[j1];
			ar[j0] += ar[j1];
			ai[j0] += ai[j1];
			ar[j1] = x0r;
			ai[j1] = x0i;
		}
	}
}

void FFT(double a[], long num)
{
	FFT_main(num, a, a + num);

	return;
}

void IFFT(double a[], long num)
{
	long cnt;

	for(cnt = 0; cnt < num; a[cnt++] *= -1);

	FFT_main(num, a, a + num);

	for(cnt = 0; cnt < num * 2; a[cnt++] /= -num);

	return;
}

void Rectangular(double a[], long num)
{
	return;
}

void Hann(double a[], long num)
{
	long cnt;

	for(cnt = 0; cnt < num; a[cnt++] *= 0.5 - 0.5 * cos(M_PI * 2 * cnt / num));

	return;
}

void Hamming(double a[], long num)
{
	long cnt;

	for(cnt = 0; cnt < num; a[cnt++] *= 0.54 - 0.46 * cos(M_PI * 2 * cnt / num));

	return;
}

void Blackman(double a[], long num)
{
	long cnt;

	for(cnt = 0; cnt < num; a[cnt++] *= 0.42 - 0.5 * cos(M_PI * 2 * cnt / num) + 0.08 * cos(M_PI * 4 * cnt / num));

	return;
}

void Bartlett(double a[], long num)
{
	long cnt;

	for(cnt = 0; cnt < num; a[cnt++] *= 1 - abs(2.0 * cnt / num - 1));

	return;
}

void Bartlett_Hann(double a[], long num)
{
	long cnt;

	for(cnt = 0; cnt < num; a[cnt++] *= 0.62 - 0.48 * abs(1.0 * cnt / num - 0.5) - 0.38 * cos(M_PI * 2 * cnt / num));

	return;
}

void Nuttall(double a[], long num)
{
	long cnt;

	for(cnt = 0; cnt < num; a[cnt++] *= 0.355768 - 0.487396 * cos(M_PI * 2 * cnt / num) + 0.144232 * cos(M_PI * 4 * cnt / num) - 0.012604 * cos(M_PI * 6 * cnt / num));

	return;
}

void Blackman_Harris(double a[], long num)
{
	long cnt;

	for(cnt = 0; cnt < num; a[cnt++] *= 0.35875 - 0.48829 * cos(M_PI * 2 * cnt / num) + 0.14128 * cos(M_PI * 4 * cnt / num) - 0.01168 * cos(M_PI * 6 * cnt / num));

	return;
}

void Blackman_Nuttall(double a[], long num)
{
	long cnt;

	for(cnt = 0; cnt < num; a[cnt++] *= 0.3635819 - 0.4891775 * cos(M_PI * 2 * cnt / num) + 0.1365995 * cos(M_PI * 4 * cnt / num) - 0.0106411 * cos(M_PI * 6 * cnt / num));

	return;
}

void flat_top(double a[], long num)
{
	long cnt;

	for(cnt = 0; cnt < num; a[cnt++] *= 1 - 1.93 * cos(M_PI * 2 * cnt / num) + 1.29 * cos(M_PI * 4 * cnt / num) - 0.388 * cos(M_PI * 6 * cnt / num) + 0.032 * cos(M_PI * 8 * cnt / num));

	return;
}

void Akaike(double a[], long num)
{
	long cnt;

	for(cnt = 0; cnt < num; a[cnt++] *= 0.625 - 0.5 * cos(M_PI * 2 * cnt / num) - 0.125 * cos(M_PI * 4 * cnt / num));

	return;
}

void Welch(double a[], long num)
{
	long cnt;

	for(cnt = 0; cnt < num; a[cnt++] *= 4.0 * cnt / num * (-1.0 * cnt / num + 1));

	return;
}

void HPF(double a[], long num, long cutoff, double pole)
{
	long cnt;

	if (cutoff <= 0)  for(cnt = 0; cnt < cutoff; a[num - ++cnt] = a[cnt] = 0.0);
	else for(cnt = 0; cnt < cutoff; a[num - ++cnt] = a[cnt] = a[cnt] * pow((double)cnt / cutoff, pole));

	return;
}

void LPF(double a[], long num, long cutoff, double pole)
{
	long cnt;

	if (cutoff <= 0) for(cnt = cutoff; cnt < num; a[num - ++cnt] = a[cnt] = 0.0);
	else for(cnt = cutoff; cnt < num; a[num - ++cnt] = a[cnt] = a[cnt] * pow((double)cutoff / cnt, pole));

	return;
}

void BPF(double a[], long num, long cutoff, double pole)
{
	//HPF(a, num, cutoff, pole);

	//LPF(a, num, cutoff, pole);

	return;
}

void BEF(double a[], long num, long cutoff, double pole)
{

	//long cnt;

	//for(cnt = 0; cnt < num; a[cnt++] = a[cnt++]);

	return;
}